/*
 * Empeg Logo file handler
 * Copyright (C) 2002 Toby Speight.  All rights reserved.
 *
 * The GIMP -- an image manipulation program
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * Based on xbm.c, X10 and X11 bitmap (XBM) loading and saving file filter for the GIMP.
 * XBM code Copyright (C) 1998 Gordon Matzigkeit.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/* Set this for debugging. */
/* #define VERBOSE 2 */
#define VERBOSE 2

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <gtk/gtk.h>

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>


#define EMPEG_WIDTH       128
#define EMPEG_HEIGHT      32
#define EMPEG_LAYER_BYTES 2048

#define MAKE_STR(x) MAKE_STR2(x)
#define MAKE_STR2(x) #x
#define EMPEG_WIDTH_STR MAKE_STR(EMPEG_WIDTH)
#define EMPEG_HEIGHT_STR MAKE_STR(EMPEG_HEIGHT)

/* i18n placeholder */
#define _(str) str


/* Data Structures */

/* In-memory representation of Empeg upload file */
typedef struct {
  guchar tag[4];
  guchar home[EMPEG_LAYER_BYTES];
  guchar car[EMPEG_LAYER_BYTES];
} EmpegData;

typedef struct _EmpegSaveVals {
  gchar type[4];                /* Empeg type identifier: "emp ", "empg", "rio ", or "rioc" */
  gint reduce_color;
  GimpConvertDitherType   dither_type;
} EmpegSaveVals;

static EmpegSaveVals xsvals = {
  "empg",                       /* type */
  FALSE,                        /* reduce_color */
  GIMP_FS_DITHER,               /* dither_type */
};


typedef struct _EmpegSaveState {
  gint is_rio;
  gint needs_reducing;
  GimpConvertDitherType   dither_type;
  gint needs_alpha_compositing;
  gint32 home_layer;
  gint32 car_layer;
} EmpegSaveState;



typedef struct _EmpegSaveDialogState {
  gint run;
  gint personality_empeg;
  gint personality_rio;
  gint dither_none;
  gint dither_fixed;
  gint dither_fslowbleed;
  gint dither_fs;
} EmpegSaveDialogState;

static EmpegSaveDialogState xsint = {
  FALSE,                        /* run */
  TRUE,                         /* personality_empeg */
  FALSE,                        /* personality_rio */
  FALSE,                        /* nodither_flag */
  FALSE,                        /* fixeddither_flag */
  FALSE,                        /* fslowbleeddither_flag */
  TRUE,                         /* fsdither_flag */
};


/* Declare some local functions.
 */
static void   query      (void);
static void   run        (char    *name,
                          int      nparams,
                          GimpParam  *param,
                          int     *nreturn_vals,
                          GimpParam **return_vals);
static gint32 load_image(char   *filename);
static gint   save_image(char   *filename,
			  gint32  image_ID);
static gint   save_dialog(gint32  drawable_ID);
static void   close_callback  (GtkWidget *widget,
			       gpointer   data);
static void   save_ok_callback     (GtkWidget *widget,
				    gpointer   data);
#if 0
static void   save_toggle_update   (GtkWidget *widget,
				    gpointer   data);
static void   comment_entry_callback(GtkWidget *widget,
				    gpointer   data);
static void   prefix_entry_callback(GtkWidget *widget,
				    gpointer   data);
#endif
static int    cmp_brightness(const void *a, const void *b);

#ifdef VERBOSE
static int verbose = VERBOSE;
#endif







static void
read_data(guchar *bytes, GimpDrawable *drawable) {

  GimpPixelRgn pixel_rgn;
  guchar *data, *p;
  int i, j, k;
  int tileheight, rowoffset;

  /* Prepare the pixel region. */
  gimp_pixel_rgn_init(&pixel_rgn, drawable, 0, 0, EMPEG_WIDTH, EMPEG_HEIGHT, TRUE, FALSE);

  /* Allocate the data. */
  tileheight = gimp_tile_height();
  data = g_malloc(EMPEG_WIDTH * tileheight);

  p = bytes;

  for (i = 0; i < EMPEG_HEIGHT; i += tileheight)
    {
      tileheight = MIN(tileheight, EMPEG_HEIGHT - i);

#ifdef VERBOSE
      if (verbose > 1)
	printf("Empeg: reading %dx(%d+%d) pixel region\n", EMPEG_WIDTH, i,
		tileheight);
#endif

      /* Parse the data from the file */
      for (j = 0; j < tileheight; j ++)
	{
	  /* Read each row. */
	  rowoffset = j * EMPEG_WIDTH;
	  for (k = 0; k < EMPEG_WIDTH; k +=2)
	    {
              guchar c = *p++;
              data[rowoffset + k]     = c & 3;
              data[rowoffset + k + 1] = c >> 4;
	    }
	}

      /* Put the data into the image. */
      gimp_progress_update((double) (i + tileheight) / (double) EMPEG_HEIGHT);
      gimp_pixel_rgn_set_rect(&pixel_rgn, data, 0, i, EMPEG_WIDTH, tileheight);
    }

  g_free(data);

  gimp_drawable_flush(drawable);
  gimp_drawable_detach(drawable);
}


static gint
load_image(char *filename)
{
  FILE *fp;

  char *name_buf;
  EmpegData *databuf;
  GimpParasite *parasite;

  gint32 image_ID;

  guchar cmap[] =
  {
    /* TODO: import this from somewhere else */
    0x00, 0x00, 0x00,		/* black */
    0x5b, 0x5b, 0x5b,
    0x73, 0x73, 0x73,
    0xff, 0xff, 0xff		/* white */
  };

  fp = fopen(filename, "rb");
  if (!fp)
    {
      printf("Empeg: cannot open \"%s\"\n", filename);
      return -1;
    }

  databuf = g_malloc(sizeof *databuf);
  if (!databuf) {
      printf("Empeg: cannot malloc %d for \"%s\"\n", sizeof *databuf, filename);
      return -1;
    }

  fread(databuf, 1, sizeof *databuf, fp);
  if (ferror(fp)) {
    printf("Empeg: cannot read \"%s\"\n", filename);
    g_free(databuf);
    fclose(fp);
    return -1;
  }

  fclose(fp);

  name_buf = g_malloc(strlen(filename) + 11);
  sprintf(name_buf, "Loading %s:", filename);
  gimp_progress_init(name_buf);
  g_free(name_buf);

  image_ID = gimp_image_new(EMPEG_WIDTH, EMPEG_HEIGHT, GIMP_INDEXED);
  gimp_image_set_filename(image_ID, filename);

  /* Store the player type in a parasite. */
  parasite = gimp_parasite_new("empeg-type", GIMP_PARASITE_PERSISTENT,
                                4, (gpointer)&(databuf->tag));
  gimp_image_parasite_attach(image_ID, parasite);
  gimp_parasite_free(parasite);

  /* Set the Empeg colormap. */
  gimp_image_set_cmap(image_ID, cmap, 4);

  if (memcmp(databuf->home, databuf->car, sizeof(databuf->home))) {
    /* Car and home images differ - use two layers */
    gint32 home_layer_ID, car_layer_ID;

    home_layer_ID = gimp_layer_new(image_ID, "home",
                                    EMPEG_WIDTH, EMPEG_HEIGHT,
                                    GIMP_INDEXED_IMAGE,
                                    100,
                                    GIMP_NORMAL_MODE);
    gimp_layer_set_visible(home_layer_ID, FALSE);
    gimp_image_add_layer(image_ID, home_layer_ID, 0);
    read_data(databuf->home, gimp_drawable_get(home_layer_ID));

    car_layer_ID = gimp_layer_new(image_ID, "car",
                                   EMPEG_WIDTH, EMPEG_HEIGHT,
                                   GIMP_INDEXED_IMAGE,
                                   100,
                                   GIMP_NORMAL_MODE);
    gimp_image_add_layer(image_ID, car_layer_ID, 0);
    read_data(databuf->car, gimp_drawable_get(car_layer_ID));
  } else {
    /* Car and home images identical - use one layer */
    gint layer_ID;

    layer_ID = gimp_layer_new(image_ID, "home",
                                    EMPEG_WIDTH, EMPEG_HEIGHT,
                                    GIMP_INDEXED_IMAGE,
                                    100,
                                    GIMP_NORMAL_MODE);
    gimp_image_add_layer(image_ID, layer_ID, 0);
    read_data(databuf->home, gimp_drawable_get(layer_ID));
  }


  g_free(databuf);
  return image_ID;
}


static void
close_callback(GtkWidget *widget,
		gpointer   data)
{
  (void)widget;
  (void)data;
  gtk_main_quit();
}

static int gtk_initialized = FALSE;


static void
error_dialog(char *msg)
{
  GtkWidget *dlg, *button, *label, *frame, *vbox;

  if (!gtk_initialized) {
      printf("empeg: %s\n", msg);
      return;
  }

  dlg = gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW(dlg), "Empeg Warning");
  gtk_window_position(GTK_WINDOW(dlg), GTK_WIN_POS_MOUSE);
  gtk_signal_connect(GTK_OBJECT(dlg), "destroy",
		      (GtkSignalFunc) close_callback,
		      dlg);
  gtk_signal_connect(GTK_OBJECT(dlg), "delete_event",
		      (GtkSignalFunc) close_callback,
		      dlg);

  /* Action area */
  button = gtk_button_new_with_label("Cancel");
  GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
  gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
			     (GtkSignalFunc) gtk_widget_destroy,
			     GTK_OBJECT(dlg));
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->action_area), button,
		      TRUE, TRUE, 0);
  gtk_widget_show(button);

      /*  the warning message  */
  frame = gtk_frame_new(NULL);
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);
  gtk_container_border_width(GTK_CONTAINER(frame), 10);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->vbox), frame, TRUE,
		      TRUE, 0);
  vbox = gtk_vbox_new(FALSE, 5);
  gtk_container_border_width(GTK_CONTAINER(vbox), 5);
  gtk_container_add(GTK_CONTAINER(frame), vbox);

  label = gtk_label_new(msg);
  gtk_box_pack_start(GTK_BOX(vbox), label, TRUE, TRUE, 0);
  gtk_widget_show(label);
  gtk_widget_show(vbox);
  gtk_widget_show(frame);
  gtk_widget_show(dlg);
  gtk_main();
  gtk_widget_destroy(GTK_WIDGET(dlg));
  gdk_flush();
}

/* Write a layer into a 2048-byte array */
static void
write_data(guchar* dest, gint drawable_ID, guchar out_map[]) {

  GimpPixelRgn pixel_rgn;
  GimpDrawable *drawable;
  int bpp;                      /* BYTES per pixel */
  
  guchar *data, *src, *last;

  drawable = gimp_drawable_get(drawable_ID);
  /* has_alpha = gimp_drawable_has_alpha(drawable_ID); */
  bpp = drawable->bpp;
  
  /* Allocate a new set of pixels.  We don't do tiling, as Empeg
     images are small.  */
  data = g_malloc(EMPEG_WIDTH * EMPEG_HEIGHT * bpp);

  gimp_pixel_rgn_init(&pixel_rgn, drawable, 0, 0,
                      EMPEG_WIDTH, EMPEG_HEIGHT,
                      FALSE, FALSE);
  gimp_pixel_rgn_get_rect(&pixel_rgn, data, 0, 0, EMPEG_WIDTH, EMPEG_HEIGHT);

  src = data;
  last = dest + EMPEG_LAYER_BYTES;

  while(dest < last) {
    *dest++ = out_map[*src] | (out_map[src[bpp]] << 4);
    src += 2*bpp;
  }

  g_free(data);
}


static gint
save_image(char   *filename,
	    gint32  image_ID)
{
  FILE *fp;

  int colors, nlayers;
  int width, height;
  int i, j;

  gint *layers;

  char  *name_buf;
  guchar *cmap;
  guchar cmap_copy[16], out_map[4];
  GimpDrawable *image;
  EmpegData *databuf;

  image = gimp_drawable_get(image_ID);
  
  cmap = gimp_image_get_cmap(image_ID, &colors);
  printf("Empeg: saving image %d\n", image_ID);

#if 0
  printf("empeg: run `gdb empeg' and `attach %d'\n", getpid());
  kill(getpid(), 19);
#endif

  width  = gimp_image_width(image_ID);
  height = gimp_image_height(image_ID);
  
  printf("Empeg: size=%dx%d\n", width, height);
  if (width != EMPEG_WIDTH || height != EMPEG_HEIGHT)
    {
      /* The image is not the right size. */
      error_dialog(
                    "The image which you are trying to save as\n"
                    "an Empeg boot logo is the wrong size.\n\n"
                    "Please resize it to " EMPEG_WIDTH_STR "x" EMPEG_HEIGHT_STR " and try again."
                    );
      return FALSE;
    }

  printf("Empeg: type=%d, colors=%d\n", gimp_drawable_type(image_ID), colors);
  xsvals.reduce_color = (gimp_drawable_type(image_ID) != GIMP_INDEXED_IMAGE || colors > 4);

  layers = gimp_image_get_layers(image_ID, &nlayers);
  printf("Image %d has %d layers\n", image_ID, nlayers);
  
  if (nlayers <= 0 || nlayers >= 3)
    {
      /* The image is not the right type. */
      error_dialog(
                    "The image which you are trying to save as an\n"
                    "Empeg logo should contain one or two layers.\n\n"
                    "Please flatten it to one or two layers and try again."
                    );
      return FALSE;
    }

  name_buf = g_malloc(strlen(filename) + 11);
  sprintf(name_buf, "Saving %s:", filename);
  gimp_progress_init(name_buf);
  g_free(name_buf);

  {
    /* Sort the colors into ascending brightness. */
    for (i = 0;  i < colors;  i++) {
      cmap_copy[i*4] = cmap[i*3];
      cmap_copy[i*4+1] = cmap[i*3+1];
      cmap_copy[i*4+2] = cmap[i*3+2];
      cmap_copy[i*4+3] = i;
    }

    qsort(cmap_copy, colors, 4, cmp_brightness);

    i = 0;
    j = 0;
    out_map[cmap_copy[j++*4+3]] = i++;
    if (colors > 3)
        out_map[cmap_copy[j++*4+3]] = i++;
    if (colors > 2)
        out_map[cmap_copy[j++*4+3]] = i++;
    if (colors > 1)             /* I can't imagine a 1-color logo, but... */
        out_map[cmap_copy[j++*4+3]] = i++;

#ifdef VERBOSE
    if (verbose > 1)
      for (i = 0;  i < 4;  i++)
        printf("out_map[%d] = %d\n", i, out_map[i]);
#endif
  }


  /* Now actually save the data. */
  fp = fopen(filename, "w");
  if (!fp) {
      printf("Empeg: cannot create \"%s\"\n", filename);
      return FALSE;
  }

  databuf = g_malloc(sizeof *databuf);
  if (!databuf) {
    printf("Empeg: cannot malloc %d for \"%s\"\n", sizeof *databuf, filename);
    fclose(fp);
    return -1;
  }

  strncpy(databuf->tag, xsvals.type, 4);

  if (nlayers==1) {
      write_data(databuf->home, layers[0], out_map);
      write_data(databuf->car,  layers[0], out_map);
  } else {
      /* nlayers==2 */
      if (!strcmp("home", gimp_layer_get_name(layers[0]))
          || !strcmp("car", gimp_layer_get_name(layers[1]))) {
          /* swap layers */
          gint tmp = layers[0];
          layers[0] = layers[1];
          layers[1] = tmp;
      }
      write_data(databuf->home, layers[1], out_map);
      write_data(databuf->car,  layers[0], out_map);
  }

  fwrite(databuf, sizeof *databuf, 1, fp);
  if (ferror(fp)) {
    printf("Empeg: cannot read \"%s\"\n", filename);
    g_free(databuf);
    fclose(fp);
    return -1;
  }

  fclose(fp);
  return TRUE;
}

static int
cmp_brightness(const void *a, const void *b) {
  const guchar *ca = a;
  const guchar *cb = b;
  /* TODO: is there a GIMP RGB->luminance function? */
  int la = 3*ca[0] + 6*ca[1] + ca[2];
  int lb = 3*cb[0] + 6*cb[1] + cb[2];
  return la - lb;
}


static gint
save_dialog(gint32  drawable_ID)
{
  GtkWidget *dlg, *frame, *vbox, *hbox, *button, *label;
  GtkWidget *vbox2;
  GtkWidget *radio;
  GSList *radio_group;

  gchar **argv;
  gint argc;

  (void)drawable_ID;

  xsint.run = FALSE;
  xsint.personality_rio   = !strncmp("rioc", xsvals.type, 4);
  xsint.personality_empeg = !xsint.personality_rio;
  xsint.dither_none       = xsvals.dither_type == GIMP_NO_DITHER;
  xsint.dither_fixed      = xsvals.dither_type == GIMP_FIXED_DITHER;
  xsint.dither_fslowbleed = xsvals.dither_type == GIMP_FSLOWBLEED_DITHER;
  xsint.dither_fs         = xsvals.dither_type == GIMP_FS_DITHER;

  argc = 1;
  argv = g_new(gchar *, 1);
  argv[0] = g_strdup("save");

  gtk_init(&argc, &argv);
  gtk_initialized = TRUE;

  gtk_rc_parse(gimp_gtkrc());
  gdk_set_use_xshm(gimp_use_xshm());

  dlg = gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW(dlg), _("Save as empeg"));
  gtk_window_position(GTK_WINDOW(dlg), GTK_WIN_POS_MOUSE);
  gtk_signal_connect(GTK_OBJECT(dlg), "destroy",
		      (GtkSignalFunc) close_callback,
		      NULL);
  gtk_signal_connect(GTK_OBJECT(dlg), "delete_event",
		      (GtkSignalFunc) close_callback,
		      dlg);

  /*  Action area  */
  button = gtk_button_new_with_label(_("OK"));
  GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
  gtk_signal_connect(GTK_OBJECT(button), "clicked",
                      (GtkSignalFunc) save_ok_callback,
                      dlg);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->action_area), button,
		      TRUE, TRUE, 0);
  gtk_widget_grab_default(button);
  gtk_widget_show(button);

  button = gtk_button_new_with_label(_("Cancel"));
  GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
  gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
			     (GtkSignalFunc) gtk_widget_destroy,
			     GTK_OBJECT(dlg));
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->action_area), button,
		      TRUE, TRUE, 0);
  gtk_widget_show(button);


  /* parameter settings */
  frame = gtk_frame_new(_("Empeg Options"));
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);
  gtk_container_border_width(GTK_CONTAINER(frame), 10);
  gtk_widget_show(frame);

  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->vbox), frame, TRUE, TRUE, 0);
  vbox = gtk_vbox_new(FALSE, 5);
  gtk_container_border_width(GTK_CONTAINER(vbox), 5);
  gtk_container_add(GTK_CONTAINER(frame), vbox);

  /* Personality(Empeg/Rio) */
  hbox = gtk_hbox_new(FALSE, 5);
  gtk_container_border_width(GTK_CONTAINER(hbox), 0);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);

  radio = gtk_radio_button_new_with_label(NULL, _("I have an Empeg Car player"));
  radio_group = gtk_radio_button_group(GTK_RADIO_BUTTON(radio));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radio), xsint.personality_empeg);
  gtk_box_pack_start(GTK_BOX(hbox), radio, FALSE, FALSE, 0);
    gtk_signal_connect(GTK_OBJECT(radio), "toggled",
                        GTK_SIGNAL_FUNC(gimp_toggle_button_update),
                        &(xsint.personality_empeg));
  gtk_widget_show(radio);

  radio = gtk_radio_button_new_with_label(radio_group, _("I have a Rio Car player"));
  radio_group = gtk_radio_button_group(GTK_RADIO_BUTTON(radio));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radio), xsint.personality_rio);
  gtk_box_pack_start(GTK_BOX(hbox), radio, FALSE, FALSE, 0);
    gtk_signal_connect(GTK_OBJECT(radio), "toggled",
                        GTK_SIGNAL_FUNC(gimp_toggle_button_update),
                        &(xsint.personality_rio));
  gtk_widget_show(radio);
  /* TODO: callbacks */
  gtk_widget_show(hbox);

#if 1
  if (xsvals.reduce_color) {
    /* Dither type. */
    frame = gtk_frame_new(_("Dither Options"));
    gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);
    gtk_container_border_width(GTK_CONTAINER(frame), 10);
    gtk_widget_show(frame);

    vbox2 = gtk_vbox_new(FALSE, 1);
    gtk_container_set_border_width(GTK_CONTAINER(vbox2), 2);
    gtk_container_add(GTK_CONTAINER(frame), vbox2);
    gtk_widget_show(vbox2);

    hbox = gtk_hbox_new(FALSE, 5);
    gtk_container_border_width(GTK_CONTAINER(hbox), 0);
    gtk_box_pack_start(GTK_BOX(vbox2), hbox, FALSE, FALSE, 0);
    label = gtk_label_new(_("Reduce to empeg palette using: "));
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
    gtk_widget_show(label);
    gtk_widget_show(hbox);

    hbox = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(vbox2), hbox, FALSE, FALSE, 0);

    radio = gtk_radio_button_new_with_label(NULL, _("No Color Dithering"));
    radio_group = gtk_radio_button_group(GTK_RADIO_BUTTON(radio));
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radio), xsint.dither_none);
    gtk_box_pack_start(GTK_BOX(hbox), radio, FALSE, FALSE, 0);
    gtk_signal_connect(GTK_OBJECT(radio), "toggled",
                        GTK_SIGNAL_FUNC(gimp_toggle_button_update),
                        &(xsint.dither_none));
    gtk_widget_show(radio);
    gtk_widget_show(hbox);

    hbox = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(vbox2), hbox, FALSE, FALSE, 0);

    radio = gtk_radio_button_new_with_label(NULL, _("Positioned Color Dithering"));
    radio_group = gtk_radio_button_group(GTK_RADIO_BUTTON(radio));
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radio), xsint.dither_fixed);
    gtk_box_pack_start(GTK_BOX(hbox), radio, FALSE, FALSE, 0);
    gtk_signal_connect(GTK_OBJECT(radio), "toggled",
                        GTK_SIGNAL_FUNC(gimp_toggle_button_update),
                        &(xsint.dither_fixed));
    gtk_widget_show(radio);
    gtk_widget_show(hbox);

    hbox = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(vbox2), hbox, FALSE, FALSE, 0);

    radio = gtk_radio_button_new_with_label(NULL, _("Floyd-Steinberg Color Dithering(Reduced Color Bleeding)"));
    radio_group = gtk_radio_button_group(GTK_RADIO_BUTTON(radio));
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radio), xsint.dither_fslowbleed);
    gtk_box_pack_start(GTK_BOX(hbox), radio, FALSE, FALSE, 0);
    gtk_signal_connect(GTK_OBJECT(radio), "toggled",
                        GTK_SIGNAL_FUNC(gimp_toggle_button_update),
                        &(xsint.dither_fslowbleed));
    gtk_widget_show(radio);
    gtk_widget_show(hbox);

    hbox = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(vbox2), hbox, FALSE, FALSE, 0);

    radio = gtk_radio_button_new_with_label(NULL, _("Floyd-Steinberg Color Dithering(Normal)"));
    radio_group = gtk_radio_button_group(GTK_RADIO_BUTTON(radio));
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radio), xsint.dither_fs);
    gtk_box_pack_start(GTK_BOX(hbox), radio, FALSE, FALSE, 0);
    gtk_signal_connect(GTK_OBJECT(radio), "toggled",
                        GTK_SIGNAL_FUNC(gimp_toggle_button_update),
                        &(xsint.dither_fs));
    gtk_widget_show(radio);
    gtk_widget_show(hbox);
  }
#endif

  /* Done. */
  gtk_widget_show(vbox);
  gtk_widget_show(dlg);

  gtk_main();
  gdk_flush();

  strncpy(xsvals.type, xsint.personality_rio ? "rioc" : "empg", 4);
  xsvals.dither_type
      = xsint.dither_none ? GIMP_NO_DITHER
      : xsint.dither_fixed ? GIMP_FIXED_DITHER
      : xsint.dither_fs ? GIMP_FS_DITHER
      : GIMP_FSLOWBLEED_DITHER;

  return xsint.run;
}

static void
save_ok_callback(GtkWidget *widget,
		  gpointer   data)
{
  (void)widget;
  xsint.run = TRUE;
  xsvals.dither_type = xsint.dither_none ? GIMP_NO_DITHER
    : xsint.dither_fixed ? GIMP_FIXED_DITHER
    : xsint.dither_fslowbleed ? GIMP_FSLOWBLEED_DITHER
    : GIMP_FS_DITHER;
   
  gtk_widget_destroy(GTK_WIDGET(data));
}

/* Run method worker functions */

static void
run_default(int        *nreturn_vals,
            GimpParam **return_vals)
{
  static GimpParam values[1];
  
  *nreturn_vals = 1;
  *return_vals  = values;
  values[0].type          = GIMP_PDB_STATUS;
  values[0].data.d_status = GIMP_PDB_CALLING_ERROR;
}

static void
run_file_empeg_load(int         nparams,
                     GimpParam  *param,
                     int        *nreturn_vals,
                     GimpParam **return_vals)
{
  static GimpParam values[2];
  GimpRunModeType  run_mode;
  gint32           image_ID;

  (void)nparams;

  run_default(nreturn_vals, return_vals); /* initialise failure case */

  run_mode = param[0].data.d_int32;

  switch (run_mode) {
  case GIMP_RUN_INTERACTIVE:
  case GIMP_RUN_WITH_LAST_VALS:
    gimp_ui_init("empeg", FALSE);
    /* TODO empeg_export_image */
    break;
  default:
    break;
  }

  image_ID = load_image(param[1].data.d_string);

  if (image_ID != -1) {
    values[0].type          = GIMP_PDB_STATUS;
    values[0].data.d_status = GIMP_PDB_SUCCESS;
    values[1].type          = GIMP_PDB_IMAGE;
    values[1].data.d_image  = image_ID;
    *nreturn_vals = 2;
    *return_vals = values;
  } 
}

static void
run_file_empeg_save(int         nparams,
                    GimpParam  *param,
                    int        *nreturn_vals,
                    GimpParam **return_vals)
{
  static GimpParam   values[1];
  GimpRunModeType    run_mode;
  GimpPDBStatusType  status   = GIMP_PDB_SUCCESS;
  gint32             image_ID;
  GimpParasite      *parasite = NULL;

  run_default(nreturn_vals, return_vals); /* initialise failure case */

  run_mode = param[0].data.d_int32;

  image_ID = param[1].data.d_int32;

  /*      int argc; */
  /*      char **argv; */

  /*      argc = 1; */
  /*      argv = g_new(gchar *, 1); */
  /*      argv[0] = g_strdup("empeg"); */

  switch (run_mode) {
  case GIMP_RUN_INTERACTIVE:
  case GIMP_RUN_WITH_LAST_VALS:
    gimp_ui_init("empeg", FALSE);
    /* TODO empeg_export_image */
    break;
  default:
    break;
  }

  switch (run_mode) {
  case GIMP_RUN_INTERACTIVE:
  case GIMP_RUN_WITH_LAST_VALS:
    /*  Possibly retrieve data  */
    gimp_get_data("file_empeg_save", &xsvals);

    /* Always override the prefix with the filename. */
    /*  	  init_prefix(param[3].data.d_string); */

    break;

  case GIMP_RUN_NONINTERACTIVE:
    /*  Make sure all the required arguments are there!  */
    if (nparams < 5)
      status = GIMP_PDB_CALLING_ERROR;
    if (status == GIMP_PDB_SUCCESS)
      {
        int i = 5;
#ifdef EMPEG_TYPE_SUPPORT
        // TODO:
        if (nparams > i)
          {
            memset(xsvals.comment, 0, sizeof(xsvals.comment));
            strncpy(xsvals.comment, param[i].data.d_string,
                     MAX_COMMENT);
          }

        i ++;
        if (nparams > i)
          xsvals.x10_format = (param[i].data.d_int32) ? TRUE : FALSE;

        i += 2;
        if (nparams > i)
          {
            /* They've asked for a hotspot. */
            xsvals.x_hot = param[i - 1].data.d_int32;
            xsvals.y_hot = param[i].data.d_int32;
          }

        i ++;
        if (nparams > i)
          {
            memset(xsvals.prefix, 0, sizeof(xsvals.prefix));
            strncpy(xsvals.prefix, param[i].data.d_string,
                     MAX_PREFIX);
          }
        else
          init_prefix(param[3].data.d_string);
#endif
        i ++;
        /* Too many arguments. */
        if (nparams > i)
          status = GIMP_PDB_CALLING_ERROR;
      }
    break;

  default:
    break;
  }

  if (run_mode == GIMP_RUN_INTERACTIVE) {
    /* Get the parasites */
    parasite = gimp_image_parasite_find(image_ID, "empeg-type");

    if (parasite) {
      gpointer data;
      gint     size;

      data = gimp_parasite_data(parasite);
      size = gimp_parasite_data_size(parasite);

      if (size==4)
        strncpy(xsvals.type, data, size);

      gimp_parasite_free(parasite);
    }

    /*  Acquire information with a dialog  */
    if (!save_dialog(image_ID)) {
      status = GIMP_PDB_CANCEL;
      values[0].data.d_status = GIMP_PDB_SUCCESS;
      return;
    }
  }

  if (xsvals.reduce_color) {
    image_ID = gimp_image_duplicate(image_ID);
    gimp_image_convert_rgb(image_ID);
    gimp_image_convert_indexed(image_ID, xsvals.dither_type,
                               GIMP_CUSTOM_PALETTE, 4,
                               FALSE, FALSE, "empeg");
  }

  *nreturn_vals = 1;
  if (save_image(param[3].data.d_string, image_ID)) {
    /*  Store xsvals data  */
    gimp_set_data("file_empeg_save", &xsvals, sizeof(xsvals));
    values[0].type          = GIMP_PDB_STATUS;
    values[0].data.d_status = GIMP_PDB_SUCCESS;
    *nreturn_vals = 1;
    *return_vals = values;
  }
    
  if (xsvals.reduce_color) {
    gimp_image_delete(image_ID);
  }

}
  

static void
run(char       *name,
     int         nparams,
     GimpParam  *param,
     int        *nreturn_vals,
     GimpParam **return_vals)
{
#ifdef VERBOSE
  if (verbose)
    printf("empeg: RUN %s\n", name);
#endif

  if (!strcmp(name, "file_empeg_load"))
    run_file_empeg_load(nparams, param, nreturn_vals, return_vals);
  else if (!strcmp(name, "file_empeg_save"))
    run_file_empeg_save(nparams, param, nreturn_vals, return_vals);
  else
    run_default(nreturn_vals, return_vals);
}

static void
query() {
  static GimpParamDef load_args[] = {
    { GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive" },
    { GIMP_PDB_STRING, "filename", "The name of the file to load" },
    { GIMP_PDB_STRING, "raw_filename", "The name entered" },
  };
  static GimpParamDef load_return_vals[] = {
    { GIMP_PDB_IMAGE, "image", "Output image" },
  };
  static int nload_args = sizeof(load_args) / sizeof(load_args[0]);
  static int nload_return_vals = sizeof(load_return_vals) / sizeof(load_return_vals[0]);


  static GimpParamDef save_args[] = {
    { GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive" },
    { GIMP_PDB_IMAGE, "image", "Input image" },
    { GIMP_PDB_DRAWABLE, "drawable", "Drawable to save" },
    { GIMP_PDB_STRING, "filename", "The name of the file to save" },
    { GIMP_PDB_STRING, "raw_filename", "The name entered" },
  } ;
  static int nsave_args = sizeof(save_args) / sizeof(save_args[0]);

  gimp_install_procedure("file_empeg_load",
                          "Load a file in Empeg Logo file format",
                          "Load a file in Empeg Logo file format.  128x32 pixels, 4 grays.",
                          "Toby Speight",
                          "Toby Speight",
                          "2002",
                          "<Load>/Empeg Logo",
			  NULL,
                          GIMP_PLUGIN,
                          nload_args, nload_return_vals,
                          load_args, load_return_vals);


  gimp_install_procedure("file_empeg_save",
                          "Save a file in Empeg Logo file format",
                          "Save a file in Empeg Logo file format.  128x32 pixels, 4 grays.",
			  "Toby Speight",
                          "Toby Speight",
                          "2002",
                          "<Save>/Empeg Logo",
			  "INDEXED",
                          GIMP_PLUGIN,
                          nsave_args, 0,
                          save_args, NULL);

  gimp_register_load_handler("file_empeg_load", "empg,empeg", "");
  gimp_register_save_handler("file_empeg_save", "empg,empeg", "");
}

/* Boilerplate to let Gimp see us. */

GimpPlugInInfo PLUG_IN_INFO =
{
  NULL,    /* init_proc */
  NULL,    /* quit_proc */
  query,   /* query_proc */
  run,     /* run_proc */
};

MAIN()




/*
Local Variables:
compile-command:"gcc -Wall -W -Wmissing-prototypes -pedantic -D__STRICT_ANSI__ -g -O -o empeg empeg.c `gtk-config --cflags --libs` -lgimp -lgimpui"
End:
*/
